--date and time subroutines
--David Money, South Carolina, U.S.A., September, 2005

global integer First2,ReadDateMode,PrintDateMode,PrintTimeMode

First2=1940 --first year that can be represented by 2 digits
ReadDateMode=1 -- 1=Y-M-D, 2=M-D-Y, 3=D-M-Y
PrintDateMode=10 --0=use ReadDateMode; 1=YYYY-MM-DD; 2=MM/DD/YYYY; 3=DD.MM.YYYY;
    -- 4=DDMONYYYY (mil.); 5=Mon. DD, YYYY; 6=Month DD, YYYY; 7=Month Dth, YYYY
    -- add 10=abbrev.day of week, date; add 20=day of week spelled out, date
PrintTimeMode=1 --1=24 hr., 2=12 hr., 3,4=include seconds

sequence Monthdays,Firstday,Day_short,Day_long,Month_short,Month_long
Monthdays={31,28,31,30,31,30,31,31,30,31,30,31}
Firstday={1,32,60,91,121,152,182,213,244,274,305,335}
Day_short={"Sun.","Mon.","Tues.","Wed.","Thur.","Fri.","Sat."}
Day_long={"Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"}
Month_short={"Jan.","Feb.","Mar.","Apr.","May","Jun.","Jul.","Aug.","Sept.",
    "Oct.","Nov.","Dec."}
Month_long={"January","February","March","April","May","June","July","August",
    "September","October","November","December"}

function val(sequence str)
    --evaluate numeric string (positive integers only)
    integer num,digit
    num=0
    for n=1 to length(str) do
	if str[n]>='0' and str[n]<='9' then
	    digit=str[n]-'0'
	    num*=10
	    num+=digit
	else exit
	end if
    end for
    return num
end function

function date_string(integer num)
    --string respresenting year, month, or day (at least 2 characters)
    sequence str
    if num<=31 then
	str=sprintf("%02d",num) --month or date; leading zero
    else
	str=sprintf("%d",num) --year; >=3 digits, no zero
    end if
    return str
end function

function ordinal(integer num)
    --returns number with th, nd, st, etc.
    sequence ending
    integer tens,units
    tens=remainder(num,100)
    units=remainder(tens,10)
    tens=floor(tens/10)
    if tens=1 then
	units=4--11 through 13 use th
    end if
    if units=1 then
	ending="st"
    elsif units=2 then
	ending="nd"
    elsif units=3 then
	ending="rd"
    else
	ending="th"
    end if
    return sprintf("%d",{num})&ending
end function

function separate(sequence str,integer sep)
    --separate string into parts between specified punctuation (may be 2 or 3)
    sequence s1,s2,s3
    integer good,i
    s1={} s2={} s3={} good=0
    i=find(sep,str)
    if i then
	good=1
	s1=str[1..i-1]
	s2=str[i+1..length(str)]
	i=find(sep,s2)
	if i then
	    s3=s2[i+1..length(s2)]
	    s2=s2[1..i-1]
	end if
    end if
    return {good,{s1,s2,s3}}
end function

function ucase(sequence str)
    --convert to uppercase
    for n=1 to length(str) do
	if str[n]>='a' and str[n]<='z' then
	    str[n]-=32
	end if
    end for
    return str
end function

function test_leap(integer yr)
    integer leap
    leap=0
    if yr/4=floor(yr/4) then
	leap=1
    end if
    if yr>1582 then
	if yr/100=floor(yr/100) then
	    if yr/400!=floor(yr/400) then
		leap=0
	    end if
	end if
    end if
    return leap
end function

function check_date(sequence ymd)
    --check for valid date
    integer good,yr,mo,da,leap,max
    good=1
    yr=ymd[1]
    mo=ymd[2]
    da=ymd[3]
    if yr<100 or mo<1 or mo>12 or da<1 then
	good=0
    end if
    leap=test_leap(yr)
    max=Monthdays[mo]
    if mo=2 then
	max+=leap
    end if
    if da>max then
	good=0
    end if
    return good
end function

function check_time(sequence hms)
    --check for valid time
    integer good,hr,mi,se
    good=1
    hr=hms[1]
    mi=hms[2]
    se=hms[3]
    if hr<0 or mi<0 or se<0 then
	good=0
    end if
    if hr>23 or mi>59 or se>59 then
	good=0
    end if
    return good
end function

function units(sequence unit,integer num)
    --add 's' when plural (or zero)
    if num=1 then
	return sprintf("%d ",{num})&unit
    else
	return sprintf("%d ",{num})&unit&'s'
    end if
end function

function today()
    --returns {year,month,day}
    sequence tmp
    tmp=date()
    return {tmp[1]+1900,tmp[2],tmp[3]}
end function

function now()
    --returns {hour,minute,second}
    sequence tmp
    tmp=date()
    return {tmp[4],tmp[5],tmp[6]}
end function

function conv_date(sequence ymd)
    --days since 99-12-31; Gregorian calendar started in 1582
    integer yr,mo,da,good,leap,xday
    good=check_date(ymd)
    if good=0 then
	return -1
    end if
    yr=ymd[1]
    mo=ymd[2]
    da=ymd[3]
    leap=test_leap(yr)
    xday=365*(yr-100)
    xday+=floor((yr-97)/4)
    if yr>1582 then
	xday-=10
	xday-=floor((yr-1501)/100)--subtract day for 1600, 1700, etc.
	xday+=floor((yr-1201)/400)--add one back for 1600, 2000, etc.
    end if
    if yr=1582 and mo>10 then
	xday-=10
    end if
    if yr=1582 and mo=10 and da>4 then-- 10 days removed from calendar
	xday-=10
    end if
    xday+=Firstday[mo]-1+da
    if leap=1 and mo>2 then
	xday+=1
    end if
    return xday
end function

function conv_time(sequence hms)
    --seconds since midnight
    integer hr,mi,se,good,xsec
    good=check_time(hms)
    if good=0 then
	return -1
    end if
    hr=hms[1]
    mi=hms[2]
    se=hms[3]
    xsec=3600*hr+60*mi+se
    return xsec
end function

function decode_date(integer xday)
    --get ymd sequence from day
    integer yr,mo,da,yrday,leap,first
    yr=floor(xday/365)+100
    yrday=remainder(xday,365)
    yrday-=floor((xday+3*365)/(4*365))
    while yrday<1 do
	yr-=1
	leap=test_leap(yr)
	yrday+=(365+leap)
    end while
    if yr=1582 and yrday>277 then--day after Oct. 4 was Oct. 15
	yrday+=10
    end if
    if yr>1582 then
	yrday+=10
	yrday+=floor((xday-1401*365)/(100*365))
	yrday-=floor((xday-1101*365)/(400*365))
    end if
    leap=test_leap(yr)
    while yrday>365+leap do
	yr+=1
	yrday-=(365+leap)
	leap=test_leap(yr)------
    end while
    for m=12 to 1 by -1 do
	first=Firstday[m]
	if m>2 then
	    first+=leap
	end if
	if first<=yrday then
	    mo=m
	    da=yrday-first+1
	    exit
	end if
    end for
    return{yr,mo,da}
end function

function decode_time(integer xsec)
    --return hms from xsec; must be from 0 to 86399
    integer hr,mi,se
    hr=floor(xsec/3600)
    mi=floor((xsec-hr*3600)/60)
    se=xsec-hr*3600-mi*60
    return {hr,mi,se}
end function

function read_date(sequence dat)
    sequence tmp,d
    integer v1,v2,v3,mode,yr,mo,da,good,cent,tmode,offset,xd
    d={}
    good=0
    yr=0 mo=0 da=0
    mode=ReadDateMode
    tmode=0
    tmp=separate(dat,'-')
    if tmp[1] then
	d=tmp[2]
    end if
    tmp=separate(dat,'/')
    if tmp[1] then
	d=tmp[2]
    end if
    tmp=separate(dat,'.')
    if tmp[1] then
	d=tmp[2]
    end if
    if length(d)=3 then
	v1=val(d[1])
	v2=val(d[2])
	v3=val(d[3])
	if mode=2 then
	    mo=v1
	    da=v2
	    yr=v3
	elsif mode=3 then
	    da=v1
	    mo=v2
	    yr=v3
	else
	    yr=v1
	    mo=v2
	    da=v3
	end if
	if yr<100 then
	    cent=floor(First2/100)
	    yr+=100*cent
	    if yr<First2 then
		yr+=100
	    end if
	end if
	good=check_date({yr,mo,da})
    else
	tmp=ucase(dat[1..3])
	if equal(tmp,"TOD") then
	    tmode=1
	    offset=0
	elsif equal(tmp,"TOM") then
	    tmode=1
	    offset=1
	elsif equal(tmp,"YES") then
	    tmode=1
	    offset=-1
	end if
	if tmode then
	    good=1
	    xd=conv_date(today())
	    xd+=offset
	    tmp=decode_date(xd)
	    yr=tmp[1] mo=tmp[2] da=tmp[3]
	end if
    end if
    return{good,{yr,mo,da}}
end function

function read_time(sequence tim)
    sequence tmp,d
    integer good,hr,mi,se,i,ap
    d={}
    good=0
    hr=0 mi=0 se=0
    ap=0
    tmp=separate(tim,':')
    if tmp[1] then
	d=tmp[2]
	hr=val(d[1])
	mi=val(d[2])
	if length(d[3]) then
	    se=val(d[3]) --remains 0 if seconds not entered
	end if
	i=find('A',ucase(tim))
	if i then
	    ap+=1
	end if
	i=find('P',ucase(tim))
	if i then
	    ap+=2
	end if
	if ap then
	    if hr=12 then
		hr=0
	    end if
	    if ap=2 then
		hr+=12
	    end if
	end if
	if ap<3 then
	    good=check_time({hr,mi,se})
	end if
    else --may be text string
	tmp=ucase(tim[1..3])
	if equal(tmp,"NOO") then
	    hr=12
	    good=1
	end if
	if equal(tmp,"MID") then --hr is already 0
	    good=1
	end if
	if equal(tmp,"NOW") then
	    tmp=now()
	    good=1
	    hr=tmp[1]
	    mi=tmp[2]
	    se=tmp[3]
	end if
    end if
    return{good,{hr,mi,se}}
end function

function read_zone(sequence tz)
    --convert time zone letter or abbreviation to signed number
    integer z,neg,good,tz1
    z=0 neg=0 good=0
    tz=ucase(tz)
    tz1=tz[1]
    if tz1='-' then
	neg=1
	if length(tz)>1 then
	    tz=tz[2..length(tz)]
	    z=val(tz)
	    if z>0 or tz[2]='0' then
		good=1
		z=-z
	    end if
	end if
    end if
    if tz1='+' then
	if length(tz)>1 then
	    tz=tz[2..length(tz)]
	    z=val(tz)
	    if z>0 or tz[2]='0' then
		good=1
	    end if
	end if
    end if
    if tz1>='0' and tz1<='9' then
	z=val(tz)
	good=1
    end if
    if length(tz)=1 then --letter code
	if tz1>='A' and tz1<='H' then
	    z=tz1-'A'+1
	elsif tz1>='J' and tz1<='M' then--I is skipped
	    z=tz1-'A'
	elsif tz1>='N' and tz1<='Y' then
	    z='M'-tz1
	elsif tz1='Z' then
	    z=0
	end if
	if z!=0 or tz1='Z' then
	    good=1
	end if
    end if
    if length(tz)>=3 then
	tz=tz[1..3]
	if equal(tz,"PST") then
	    z=-8 good=1
	elsif equal(tz,"PDT") or equal(tz,"PDS") or equal(tz,"MST")  then
	    z=-7 good=1
	elsif equal(tz,"MDT") or equal(tz,"MDS") or equal(tz,"CST") then
	    z=-6 good=1
	elsif equal(tz,"CDT") or equal(tz,"CDS") or equal(tz,"EST") then
	    z=-5 good=1
	elsif equal(tz,"EDT") or equal(tz,"EDS") or equal(tz,"AST") then
	    z=-4 good=1
	elsif equal(tz,"ADT") or equal(tz,"ADS") then
	    z=-3 good=1
	elsif equal(tz,"GMT") or equal(tz,"UTC") or equal(tz,"WET") then
	    z=0 good=1
	elsif equal(tz,"BST") or equal(tz,"WES") or equal(tz,"CET") then
	    z=1 good=1
	elsif equal(tz,"CES") or equal(tz,"EET") then
	    z=2 good=1
	elsif equal(tz,"JST") then
	    z=9 good=1
	elsif equal(tz,"GST") then
	    z=10 good=1
	elsif equal(tz,"NZS") then
	    z=12 good=1
	elsif equal(tz,"NZD") then
	    z=13 good=1
	end if
    end if
    return {good,z}
end function    

function dayofweek(sequence ymd)
    integer xdate,dow
    xdate=conv_date(ymd)
    dow=remainder(xdate+2,7)+1 --1=Sunday
    if xdate=0 then
	dow=0
    end if
    return dow
end function

function make_date(sequence ymd)
    --format date string for display
    sequence mstring,dstring
    integer yr,mo,da,sep,datmode,dowmode,dow
    yr=ymd[1]
    mo=ymd[2]
    da=ymd[3]
    datmode=remainder(PrintDateMode,10)
    if datmode=0 then
	datmode=ReadDateMode
    end if
    dowmode=floor(PrintDateMode/10)
    if datmode<=1 or datmode>7 then
	sep='-'
	dstring=date_string(yr)&sep&date_string(mo)&sep&date_string(da)
    elsif datmode=2 then
	sep='/'
	dstring=date_string(mo)&sep&date_string(da)&sep&date_string(yr)
    elsif datmode=3 then
	sep='.'
	dstring=date_string(da)&sep&date_string(mo)&sep&date_string(yr)
    elsif datmode=4 then
	mstring=Month_short[mo]
	mstring=ucase(mstring[1..3])
	dstring=date_string(da)&mstring&date_string(yr)
    elsif datmode=5 then
	mstring=Month_short[mo]
	dstring=mstring&' '&date_string(da)&", "&date_string(yr)
    elsif datmode=6 then
	mstring=Month_long[mo]
	dstring=mstring&' '&date_string(da)&", "&date_string(yr)
    elsif datmode=7 then
	mstring=Month_long[mo]
	dstring=mstring&' '&ordinal(da)&", "&date_string(yr)
    end if
    if dowmode>0 then
	dow=dayofweek(ymd)
    end if
    if dowmode=1 then
	dstring=Day_short[dow]&", "&dstring
    end if
    if dowmode>1 then
	dstring=Day_long[dow]&", "&dstring
    end if
    return dstring
end function

function make_time(sequence hms)
    sequence d1,d2,d3,ap
    integer hr,min,sec,mode
    ap=" am"
    hr=hms[1]
    min=hms[2]
    sec=hms[3]
    mode=PrintTimeMode
    if mode<=1 or mode=3 or mode>4then
	d1=date_string(hr)
	d2=date_string(min)
	d3=date_string(sec)
	if mode<=1 then
	    return d1&':'&d2
	else
	    return d1&':'&d2&':'&d3
	end if
    else
	--trace(1)
	if hr>=12 then
	    ap=" pm"
	    hr-=12
	end if
	if hr=0 then
	    hr=12
	end if
	d1=date_string(hr)
	d2=date_string(min)
	d3=date_string(sec)
	if mode=2 then
	    return d1&':'&d2&ap
	else
	    return d1&':'&d2&':'&d3&ap
	end if
    end if
end function

global function show_date(sequence dat)
    --return text string for date
    sequence ymd
    ymd=read_date(dat)
    if ymd[1]=0 then
	return "Invalid date"
    else
	return make_date(ymd[2])
    end if
end function

global function show_time(sequence tim)
    --return text string for time
    sequence hms
    hms=read_time(tim)
    if hms[1]=0 then
	return "Invalid time"
    else
	return make_time(hms[2])
    end if
end function

global function show_new_date(sequence dat,integer days)
    --return string for date before or after given date
    sequence ymd
    integer xd
    ymd=read_date(dat)
    if ymd[1]=0 then
	return "Invalid date"
    else
	xd=conv_date(ymd[2])
	xd+=days
	if xd<1 then
	    return "New date out of range"
	end if
	ymd=decode_date(xd)
	return make_date(ymd)
    end if
end function

global function show_new_time(sequence tim,integer secs)
    --return string for time before or after given time
    sequence hms,dstring
    integer xs,days
    dstring={} days=0
    hms=read_time(tim)
    if hms[1]=0 then
	return "Invalid time"
    else
	xs=conv_time(hms[2])
	xs+=secs
	while xs<0 do
	    xs+=86400
	    days-=1
	end while
	while xs>=86400 do
	    xs-=86400
	    days+=1
	end while
	if days<-1 then
	    dstring=" ("&sprintf("%d",-days)&" days earlier)"
	elsif days=-1 then
	    dstring=" (previous day)"
	elsif days=1 then
	    dstring=" (next day)"
	elsif days>1 then
	    dstring=" ("&sprintf("%d",days)&" days later)"
	end if
	hms=decode_time(xs)
	return make_time(hms)&dstring
    end if
end function

global function show_new_date_time(sequence dat,sequence tim,integer days,integer secs)
    --gives date and time before or after by number of days, seconds
    sequence ymd,hms
    integer xd,xs
    ymd=read_date(dat)
    if ymd[1]=0 then
	return "Invalid date"
    else
	xd=conv_date(ymd[2])
	xd+=days
    end if
    hms=read_time(tim)
    if hms[1]=0 then
	return "Invalid time"
    else
	xs=conv_time(hms[2])
	xs+=secs
    end if
    xd+=floor(xs/86400)
    xs=remainder(xs,86400)
    if xs<0 then
	xs+=86400
    end if
    if xd<1 then
	return "New date out of range"
    end if
    ymd=decode_date(xd)
    hms=decode_time(xs)
    return make_time(hms)&" on "&make_date(ymd)
end function

global function different_zone(sequence tim,sequence tz1,sequence tz2)
    --gives time in a different time zone (-12 to +12, A to Z, cst, edt, utc)
    sequence zone
    integer z1,z2
    zone=read_zone(tz1)
    if zone[1]=0 then
	return "Invalid time zone #1"
    else
	z1=zone[2]
    end if
    zone=read_zone(tz2)
    if zone[1]=0 then
	return "Invalid time zone #2"
    else
	z2=zone[2]
    end if
    return show_new_time(tim,3600*(z2-z1))
end function

global function interval_days(sequence dat1,sequence dat2)
    --gives number of days between 2 dates
    sequence tmp,ymd1,ymd2,xstr,ystr,mstr,dstr
    integer xd1,xd2,y1,m1,y,m,d,leap,days
    ymd1=read_date(dat1)
    ymd2=read_date(dat2)
    if ymd1[1]=0 then
	return "Invalid date #1"
    end if
    if ymd2[1]=0 then
	return "Invalid date #2"
    end if
    ymd1=ymd1[2] ymd2=ymd2[2]
    xd1=conv_date(ymd1)
    xd2=conv_date(ymd2)
    if xd1>xd2 then--swap if first is later than second
	tmp=ymd1
	ymd1=ymd2
	ymd2=tmp
    end if
    y1=ymd1[1]
    y=ymd2[1]-y1
    m1=ymd1[2]
    m=ymd2[2]-m1
    if m<0 then
	m+=12
	y-=1
    end if
    d=ymd2[3]-ymd1[3]
    if d<0 then
	leap=test_leap(y1)
	days=Monthdays[m1]
	if m1=2 then
	    days+=leap
	end if
	d+=days
	m-=1
	if m<0 then
	    m+=12
	    y-=1
	end if
    end if
    xstr=units("day",xd2-xd1)
    ystr=units("year",y)
    mstr=units("month",m)
    dstr=units("day",d)
    if y=0 and m=0 then
	return xstr
    end if
    if y=0 then
	return xstr&" ("&mstr&" and "&dstr&")"
    else
	return xstr&" ("&ystr&", "&mstr&", "&dstr&")"
    end if
    
end function

global function interval_secs(sequence tim1,sequence tim2)
    --gives seconds between 2 times
    sequence hms1,hms2,tmp,xstr,hstr,mstr,sstr
    integer xs1,xs2,h,m,s
    hms1=read_time(tim1)
    if hms1[1]=0 then
	return "Invalid time #1"
    end if
    hms2=read_time(tim2)
    if hms2[1]=0 then
	return "Invalid time #2"
    end if
    hms1=hms1[2] hms2=hms2[2]
    xs1=conv_time(hms1)
    xs2=conv_time(hms2)
    if xs1>xs2 then
	tmp=hms1
	hms1=hms2
	hms2=tmp
    end if
    h=hms2[1]-hms1[1]
    m=hms2[2]-hms1[2]
    if m<0 then
	m+=60
	h-=1
    end if
    s=hms2[3]-hms1[3]
    if s<0 then
	s+=60
	m-=1
	if m<0 then
	    m+=60
	    h-=1
	end if
    end if
    xstr=units("second",xs2-xs1)
    hstr=units("hour",h)
    mstr=units("minute",m)
    sstr=units("second",s)
    if h=0 and m=0 then
	return xstr
    end if
    if h=0 then
	return xstr&" ("&mstr&" and "&sstr&")"
    else
	return xstr&" ("&hstr&", "&mstr&", "&sstr&")"
    end if
end function

global function interval_days_secs(sequence dat1,sequence tim1,sequence dat2,sequence tim2)
    --time between 2 events in days and seconds
    sequence ymd1,ymd2,hms1,hms2,tmpd,tmps,xdstr,yrstr,mostr,dastr,xsstr,hrstr,
    mistr,sestr
    integer xd1,xd2,xs1,xs2,yp,mp,yr,mo,da,leap,days,hr,mi,se
    ymd1=read_date(dat1)
    if ymd1[1]=0 then
	return "Invalid date #1"
    end if
    ymd2=read_date(dat2)
    if ymd2[1]=0 then
	return "Invalid date #2"
    end if
    hms1=read_time(tim1)
    if hms1[1]=0 then
	return "Invalid time #1"
    end if
    hms2=read_time(tim2)
    if hms2[1]=0 then
	return "Invalid time #2"
    end if
    ymd1=ymd1[2] ymd2=ymd2[2]
    hms1=hms1[2] hms2=hms2[2]
    xd1=conv_date(ymd1)
    xd2=conv_date(ymd2)
    xs1=conv_time(hms1)
    xs2=conv_time(hms2)
    if xd1>xd2 or (xd1=xd2 and xs1>xs2) then--swap if first is later
	tmpd=ymd1 tmps=hms1
	ymd1=ymd2 hms1=hms2
	ymd2=tmpd hms2=tmps
	if xs2>xs1 then
	    xs1+=86400
	    xd1-=1
	end if
    else
	if xs1>xs2 then
	    xs2+=86400
	    xd2-=1
	end if
    end if
    yp=ymd2[1]
    yr=ymd2[1]-ymd1[1]
    mp=ymd2[2]-1--month previous to later date; count days from there
    if mp<1 then
	mp+=12
	yp-=1
    end if
    mo=ymd2[2]-ymd1[2]
    if mo<0 then
	mo+=12
	yr-=1
    end if
    da=ymd2[3]-ymd1[3]
    if da<0 then
	leap=test_leap(yp)
	days=Monthdays[mp]
	if mp=2 then
	    days+=leap
	end if
	da+=days
	mo-=1
	if mo<0 then
	    mo+=12
	    yr-=1
	end if
    end if
    hr=hms2[1]-hms1[1]
    mi=hms2[2]-hms1[2]
    if mi<0 then
	mi+=60
	hr-=1
    end if
    se=hms2[3]-hms1[3]
    if se<0 then
	se+=60
	mi-=1
	if mi<0 then
	    mi+=60
	    hr-=1
	end if
    end if
    if hr<0 then
	hr+=24
	da-=1
	if da<0 then
	    leap=test_leap(yp)
	    days=Monthdays[mp]
	    if mp=2 then
		days+=leap
	    end if
	    da+=days
	    mo-=1
	    if mo<0 then
		mo+=12
		yr-=1
	    end if
	end if
    end if
    xdstr=units("day",xd2-xd1)
    yrstr=units("year",yr)
    mostr=units("month",mo)
    dastr=units("day",da)
    xsstr=units("second",xs2-xs1)
    hrstr=units("hour",hr)
    mistr=units("minute",mi)
    sestr=units("second",se)
    if yr=0 and mo=0 and da=0 and hr=0 and mi=0 then
	return xsstr
    end if
    if yr=0 and mo=0 and da=0 and hr=0 then
	return xsstr&" ("&mistr&" and "&sestr&")"
    end if
    if yr=0 and mo=0 and da=0 then
	return xsstr&" ("&hrstr&", "&mistr&", "&sestr&")"
    end if
    if yr=0 and mo=0 then
	return xdstr&" and "&xsstr&" ("&dastr&", "&hrstr&", "&mistr&", "&sestr&")"
    end if
    if yr=0 then
	return xdstr&" and "&xsstr&" ("&mostr&", "&dastr&", "&hrstr&", "&mistr&
	", "&sestr&")"
    else
	return xdstr&" and "&xsstr&" ("&yrstr&", "&mostr&", "&dastr&", "&hrstr&
	", "&mistr&", "&sestr&")"
    end if
end function

global function calendar(integer yr,integer mo)
    --prints out a calendar
    sequence mline,dow,dline,noday,margin,wline,ymd,output
    integer cent,max,leap,len,xday,mday,wday,mday2
    if yr<0 then
	return "Invalid year"
    end if
    if yr<=99 then
	cent=floor(First2/100)
	yr+=100*cent
	if yr<First2 then
	    yr+=100
	end if
    end if
    if mo<1 or mo>12 then
	return "Invalid month"
    end if
    leap=test_leap(yr)
    max=Monthdays[mo]
    if mo=2 then
	max+=leap
    end if
    mline=Month_long[mo]& sprintf(", %d",{yr})
    len=length(mline)
    if len<33 then
	mline=repeat(' ',floor((33-len)/2))& mline
    end if
    dow=repeat(repeat(' ',5),7)
    for n=1 to 7 do
	dow[n][1..3]=ucase(Day_long[n][1..3])
    end for
    dline={}
    for n=1 to 7 do
	dline&=dow[n]
    end for
    margin=repeat(' ',4) --left margin; change if necessary
    output={}
    mline=margin & mline
    dline=margin & dline
    output="\n" & mline & "\n\n" & dline & "\n\n"
    noday=repeat(' ',5)
    xday=conv_date({yr,mo,1})
    wday=remainder(xday+2,7)+1
    wline={}
    if wday!=1 then
	for n=1 to wday-1 do
	    wline&=noday
	end for
    end if
    --each day is checked so it will work in 1582; only necessary for 1st week
    for n=wday to 7 do
	ymd=decode_date(xday+n-wday)
	mday=ymd[3]
	wline&=sprintf("%2d   ",{mday})
    end for
    output&=margin & wline & "\n"
    mday2=mday+1--1st day of 2nd week
    for row=0 to 4 do
	if mday>=max then
	    exit
	end if
	wline={}
	for col=0 to 6 do
	    mday=mday2+7*row+col
	    if mday<=max then
		wline&=sprintf("%2d   ",{mday})
	    end if
	end for
	output=output & margin & wline &"\n"
    end for
    return output
end function

global function current_calendar()
    --print calendar for this month
    sequence ymd
    ymd=today()
    return calendar(ymd[1],ymd[2])
end function
